"""
PyLuogu
=======

A model-based Python implement for Luogu API client

洛谷 API 客户端基于模型的 Python 实现
"""

from .exceptions import AccessDeniedHttpException, HttpException, NotFoundHttpException
from .models.main import Problem, User
from .models.paste import Paste
from .session import Session

__version__ = "0.1.0"

__all__ = (
    "AccessDeniedHttpException",
    "HttpException",
    "NotFoundHttpException",
    "Paste",
    "Problem",
    "Session",
    "User",
)



import requests

from ..constants import USER_AGENT
from ..exceptions import AccessDeniedHttpException, HttpException, NotFoundHttpException
from ..utils import dict_without_underscores, get_csrf_token


class Model:
    _session = requests.Session()
    _session.headers["User-Agent"] = USER_AGENT

    @classmethod
    def _get(cls, url: str, params: dict = None, check: bool = True) -> "dict[str]":
        r = cls._session.get(
            url,
            params=params,
            headers={"X-Luogu-Type": "content-only"},
        )
        r.raise_for_status()
        data = r.json()
        if check:
            if data["code"] == 404:
                raise NotFoundHttpException(data["currentData"]["errorMessage"])
            elif data["code"] == 403:
                raise AccessDeniedHttpException(data["currentData"]["errorMessage"])
            elif data["code"] >= 400:
                raise HttpException(data["currentData"]["errorMessage"])
        return data

    @classmethod
    def _post(cls, url: str, data: dict = None) -> "dict[str]":
        r = cls._session.post(
            url,
            json=data,
            headers={
                "x-csrf-token": get_csrf_token(cls._session),
            },
        )
        r.raise_for_status()
        return r.json()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.id})"

    def __eq__(self, other):
        if hasattr(self, "_current_data") and hasattr(other, "_current_data"):
            return self._current_data == other._current_data
        if type(other) is type(self):
            return dict_without_underscores(self.__dict__) == dict_without_underscores(
                other.__dict__
            )
        else:
            super().__eq__(other)



from datetime import datetime

from ..utils import cached_method
from . import Model
from .main import User


class Paste(Model):
    """剪贴板

    :param str id: 剪贴板 ID

    :var str data: 内容
    :var str id: 剪贴板 ID
    :var User user: 用户
    :var datetime.datetime time: 时间
    :var bool public: 是否公开
    """

    def __init__(self, id: str) -> None:
        self._current_data: dict[str] = self._get(
            f"https://www.luogu.com.cn/paste/{id}"
        )["currentData"]

        paste: dict[str] = self._current_data["paste"]
        self.data: str = paste["data"]
        self.id: str = paste["id"]
        self._user: dict[str] = paste["user"]
        self.time = datetime.fromtimestamp(paste["time"])
        self.public: bool = paste["public"]

    @property
    @cached_method
    def user(self) -> User:
        return User(self._user["uid"])

    def delete(self) -> str:
        """删除剪贴板

        :returns: 剪贴板 ID
        :rtype: str
        """
        return self._post(f"https://www.luogu.com.cn/paste/delete/{self.id}")["id"]

    def edit(self, data: str = None, public: bool = None):
        """编辑剪贴板

        :param str data: 剪贴板内容
        :param bool public: 是否公开

        :returns: 剪贴板 ID
        :rtype: str
        """
        r = self._post(
            f"https://www.luogu.com.cn/paste/edit/{self.id}",
            {"data": data, "public": public},
        )
        if data is not None:
            self.data = data
        if public is not None:
            self.public = public
        return r["id"]

    @classmethod
    def new(cls, data: str, public: bool = None) -> "Paste":
        """新建剪贴板

        :param str data: 剪贴板内容
        :param bool public: 值为真时表示公开剪贴板，否则表示私有剪贴板

        :rtype: Session.Paste
        """
        r = cls._post(
            "https://www.luogu.com.cn/paste/new",
            {"data": data, "public": public},
        )
        return cls(r["id"])



from html.parser import HTMLParser

import requests


def dict_without_underscores(d: dict):
    return dict(filter(lambda i: not i[0].startswith("_"), d.items()))


def get_csrf_token(
    session: requests.Session, url: str = "https://www.luogu.com.cn/"
) -> str:
    class HTMLCSRFTokenParser(HTMLParser):
        def handle_starttag(self, tag, attrs):
            attrs = dict(attrs)
            try:
                if tag == "meta" and attrs["name"] == "csrf-token":
                    raise StopIteration(attrs["content"])
            except KeyError:
                pass

    r = session.get(url)
    r.raise_for_status()
    try:
        HTMLCSRFTokenParser().feed(r.text)
    except StopIteration as csrf_token:
        return str(csrf_token)


def cached_method(func):
    def wrapper(self, *args):
        if not hasattr(self, "__cache"):
            self.__cache = {}
        cache = self.__cache
        if args not in cache:
            cache[args] = func(self, *args)
        return cache[args]

    return wrapper


class LazyList(list):
    def __init__(self, model, args):
        super().__init__(args)
        self._model = model

    @cached_method
    def __getitem__(self, index):
        return self._model(super().__getitem__(index))

    def __iter__(self):
        for i in super().__iter__():
            yield self._model(i)

    def __repr__(self) -> str:
        return (
            "["
            + ",\n ".join([f"{self._model.__name__}({i})" for i in list.__iter__(self)])
            + "]"
        )



from http.cookies import SimpleCookie
from io import BytesIO

import requests

from .constants import USER_AGENT
from .models.main import Problem, User
from .models.paste import Paste
from .utils import get_csrf_token


class Session:
    """会话

    :param cookies: Cookies
    :type cookies: str | dict[str, str] | None

    :var requests.cookies.RequestsCookieJar cookies: Cookies
    :var requests.Session session: 会话
    """

    def __init__(self, cookies: "str | dict[str, str] | None" = None) -> None:
        self.cookies = requests.cookies.cookiejar_from_dict(
            {k: v.value for k, v in SimpleCookie(cookies).items()}
        )
        self.session = requests.Session()
        self.session.headers["User-Agent"] = USER_AGENT
        self.session.headers["referer"] = "http://www.luogu.com.cn/"
        self.session.cookies = self.cookies
        self.Paste._session = self.session
        self.Problem._session = self.session
        self.User._session = self.session

    class Paste(Paste):
        pass

    class Problem(Problem):
        pass

    class User(User):
        pass

    def captcha(self, show: bool = True) -> bytes:
        """获取验证码

        :param bool show:
            值为真时使用 :meth:`PIL.Image.Image.show` 显示验证码；否则仅返回图片的二进制数据

        :rtype: bytes
        """
        r = self.session.get("https://www.luogu.com.cn/api/verify/captcha")
        r.raise_for_status()
        if show:
            from PIL import Image

            Image.open(BytesIO(r.content)).show(title="CAPTCHA")
        return r.content

    def login(self, username: str, password: str, captcha: str) -> "dict[str]":
        """登录

        :param str username: 用户名
        :param str password: 密码
        :param str captcha: 验证码

        :rtype: dict[str]
        """

        r = self.session.post(
            "https://www.luogu.com.cn/api/auth/userPassLogin",
            headers={
                "x-csrf-token": get_csrf_token(
                    self.session, "https://www.luogu.com.cn/auth/login"
                ),
            },
            json={
                "username": username,
                "password": password,
                "captcha": captcha,
            },
        )
        r.raise_for_status()
        return r.json()

    def logout(self) -> "dict[str, bool]":
        """登出

        :rtype: dict[str, bool]
        """
        r = self.session.post(
            "https://www.luogu.com.cn/api/auth/logout",
            headers={"x-csrf-token": get_csrf_token(self.session)},
        )
        r.raise_for_status()
        return r.json()
